from abc import ABC, abstractmethod
from django.urls import reverse
from django.conf import settings
from rest_framework.exceptions import NotAuthenticated, AuthenticationFailed, APIException, ValidationError
from gitlab import Gitlab, GitlabGetError, GitlabCreateError
from arkindex.documents.models import Corpus
import urllib.parse
import base64
import uuid
import logging

logger = logging.getLogger(__name__)


class GitProvider(ABC):

    display_name = None

    def __init__(self, credentials=None, url=None):
        if credentials:
            from arkindex.users.models import OAuthCredentials
            assert isinstance(credentials, OAuthCredentials)
            self.url = credentials.provider_url
        if url:
            self.url = url
        self.credentials = credentials

    @abstractmethod
    def list_repos(self, query=None):
        """
        List all repositories or filter with a search query.
        """

    @abstractmethod
    def get_latest_commit_sha(self, repo):
        """
        Retrieve latest commit sha from the default branch on a given repository.
        """

    @abstractmethod
    def create_repo(self, **kwargs):
        """
        Create a Repository instance from an external repository
        """

    def update_or_create_ref(self, repo, revision, name, type):
        """
        Update or create a GitRef on a given repository:
        - If a GitRef with given name already exists on the given repository, we only update the
        linked revision with the given one
        - Else a GitRef is created on the given repository thanks to given name, type and revision
        """
        from arkindex.dataimport.models import GitRef
        try:
            ref = repo.refs.get(name=name)
            ref.revision = revision
            ref.save()
        except GitRef.DoesNotExist:
            repo.refs.create(name=name, type=type, revision=revision)

    def get_or_create_revision(self, repo, sha, save=True):
        from arkindex.dataimport.models import Revision
        try:
            return self.get_revision(repo, sha), False
        except Revision.DoesNotExist:
            return self.create_revision(repo, sha, save=save), True

    def get_revision(self, repo, sha):
        return repo.revisions.get(hash=sha)

    @abstractmethod
    def create_revision(self, repo, sha, save=True):
        """
        Create a Revision instance for a given commit hash of a given repository.
        """

    @abstractmethod
    def handle_webhook(self, repo, request):
        """
        Handle a webhook event on a given repository.
        """


class GitLabProvider(GitProvider):

    display_name = "GitLab"
    url = 'https://gitlab.com'

    def _try_get_project(self, gl, id):
        try:
            return gl.projects.get(id)
        except GitlabGetError as e:
            raise APIException("Error while fetching GitLab project: {}".format(str(e)))

    def list_repos(self, query=None):
        if not self.credentials:
            raise NotAuthenticated()
        gl = Gitlab(self.url, oauth_token=self.credentials.token)
        return gl.projects.list(membership=True, search=query)

    def get_latest_commit_sha(self, repo):
        gl = Gitlab(self.url, oauth_token=repo.credentials.token)
        # Since ref_name isn't specified, we will use the repository default branch
        # See : https://docs.gitlab.com/ee/api/commits.html#list-repository-commits
        project = self._try_get_project(gl, urllib.parse.urlsplit(repo.url).path.strip('/'))

        return project.commits.list()[0].id

    def create_repo(self, id=None, corpus=None, request=None, **kwargs):
        assert isinstance(corpus, Corpus)
        if not self.credentials and request:
            raise NotAuthenticated()
        gl = Gitlab(self.url, oauth_token=self.credentials.token)
        project = self._try_get_project(gl, int(id))

        from arkindex.dataimport.models import Repository
        if Repository.objects.filter(url=project.web_url).exists():
            raise ValidationError("A repository with this URL already exists")

        repo = self.credentials.repos.create(
            corpus=corpus,
            url=project.web_url,
            hook_token=base64.b64encode(uuid.uuid4().bytes).decode('utf-8'),
            provider_name=self.__class__.__name__,
        )

        try:
            url = reverse('api:import-hook', kwargs={'pk': repo.id})
            if settings.BACKEND_PUBLIC_URL_OAUTH:
                url = urllib.parse.urljoin(settings.BACKEND_PUBLIC_URL_OAUTH, url)
            else:
                url = request.build_absolute_uri(url)

            project.hooks.create({
                'url': url,
                'push_events': True,
                'token': repo.hook_token,
            })
        except GitlabCreateError as e:
            raise APIException("Error while creating GitLab hook: {}".format(str(e)))

        return repo

    def create_revision(self, repo, sha, save=True):
        from arkindex.dataimport.models import Revision, GitRefType

        gl = Gitlab(self.url, oauth_token=repo.credentials.token)
        project = self._try_get_project(gl, urllib.parse.urlsplit(repo.url).path.strip('/'))
        commit = project.commits.get(sha)

        rev = Revision(
            repo=repo,
            hash=sha,
            message=commit.message,
            author=commit.author_name,
        )
        if save:
            rev.save()
            for ref in commit.refs():
                try:
                    ref_type = GitRefType(ref['type'])
                    self.update_or_create_ref(repo, rev, ref['name'], ref_type)
                except ValueError:
                    logger.warning(f'Git reference with type {ref["type"]} was ignored during revision creation')
                    continue

        return rev

    def handle_webhook(self, repo, request):
        from arkindex_common.enums import DataImportMode

        if 'HTTP_X_GITLAB_EVENT' not in request.META:
            raise ValidationError("Missing GitLab event type")
        if request.META['HTTP_X_GITLAB_EVENT'] != 'Push Hook':
            raise ValidationError("Unsupported GitLab event type")

        if 'HTTP_X_GITLAB_TOKEN' not in request.META:
            raise NotAuthenticated("Missing GitLab secret token")
        if request.META['HTTP_X_GITLAB_TOKEN'] != repo.hook_token:
            raise AuthenticationFailed("Invalid GitLab secret token")

        if not isinstance(request.data, dict) or 'object_kind' not in request.data:
            raise ValidationError('Bad payload format')

        if not request.data['object_kind'] == 'push':
            raise ValidationError("Unsupported GitLab event type")

        sha = request.data.get('checkout_sha')
        if not sha:
            raise ValidationError('Missing checkout SHA')

        # Already took care of this event
        if repo.revisions.filter(hash=sha).exists():
            return

        rev = self.create_revision(repo, sha)

        di = repo.corpus.imports.create(
            creator=repo.credentials.user,
            mode=DataImportMode.Repository,
            revision=rev,
            payload={
                'repo_id': str(repo.id),
                'sha': sha,
            },
        )
        di.start()


git_providers = [
    GitLabProvider,
]
oauth_to_git = {
    "gitlab": GitLabProvider,
}


def get_provider(name):
    return next(filter(lambda p: p.__name__ == name, git_providers), None)


def from_oauth(name):
    return oauth_to_git.get(name)
